import logging
import time
from itertools import pairwise, chain
from typing import TypeVar

from th2_common.schema.event.event_batch_router import EventBatchRouter
from th2_common.schema.factory.common_factory import CommonFactory
from th2_common_utils import create_event, create_event_id
from th2_grpc_common.common_pb2 import EventBatch

from run_one.action_handlers.abstract_action_handler import AbstractActionHandler
from run_one.action_handlers.context import Context
from run_one.processors.abstract_processor import AbstractProcessor
from run_one.util.config import Config
from run_one.util.util import Action


class Th2ProcessorConfig:
    def __init__(self, **kwargs) -> None:

        self.th2_configs = 'th2_configs'
        if 'th2_configs' in kwargs:
            self.th2_configs = kwargs['th2_configs']

        self.book = 'book'
        if 'book' in kwargs:
            self.book = kwargs['book']

        self.scope = 'scope'
        if 'scope' in kwargs:
            self.scope = kwargs['scope']

        self.use_place_method = False
        if 'use_place_method' in kwargs:
            self.use_place_method = bool(kwargs['use_place_method'])

        self.key_fields = []
        if 'key_fields' in kwargs:
            self.key_fields = kwargs['key_fields']

        self.sleep = 0
        if 'sleep' in kwargs:
            self.sleep = int(kwargs['sleep'])


class Th2Processor(AbstractProcessor):
    def __init__(self, config: Config, processor_config_class=Th2ProcessorConfig):
        self._config = processor_config_class(**config.processor_config)

        self._common_factory = CommonFactory(config_path=self._config.th2_configs)

        self._event_router: EventBatchRouter = self._common_factory.event_batch_router  # type: ignore
        self.root_event_id = create_event_id(book_name=self._config.book, scope=self._config.scope)
        self.root_event = EventBatch(events=[create_event(name='Run One Root Event',
                                                          event_id=self.root_event_id)])
        self._event_router.send(self.root_event)
        Context.set('root_event_id', self.root_event_id)

        self._grpc_router = self._common_factory.grpc_router

        HandlerType = TypeVar('HandlerType', bound=AbstractActionHandler)
        action_handlers = self.load_action_handlers(config.processed_actions)
        class_instance_mapping = {x: x(self._config, self._grpc_router) for x in set(action_handlers.values())}
        self.processed_actions: dict[str, HandlerType] = {action: class_instance_mapping[action_handler]
                                                          for action, action_handler in action_handlers.items()}

        self.logger = logging.getLogger()

    def process(self, test_cases: dict[str, list[Action]]):

        for test_case_name, actions in test_cases.items():

            logging.info(f'Processing {test_case_name} test case')

            test_case_root_event_id = create_event_id(book_name=self._config.book, scope=self._config.scope)
            test_case_event_batch = EventBatch(events=[create_event(name=test_case_name,
                                                                    event_id=test_case_root_event_id,
                                                                    parent_id=self.root_event_id)])
            self._event_router.send(test_case_event_batch)
            Context.set('parent_event_id', test_case_root_event_id)

            for previous_action, current_action in pairwise(chain([None], actions)):

                current_action_type = current_action.extra_data['Action']

                self.logger.info(f'Processing {current_action_type} action '
                                 f'with ID = {current_action.extra_data.get("ID", "empty")}, '
                                 f'message type = {current_action.extra_data.get("MessageType", "none")},'
                                 f'description = {current_action.extra_data.get("Description", "empty")}')

                current_action_handler = self.processed_actions.get(current_action_type)

                if previous_action is not None:
                    previous_action_handler = self.processed_actions.get(previous_action.extra_data['Action'])
                    if current_action_handler is not previous_action_handler:
                        previous_action_handler.on_action_change(previous_action, current_action)

                if current_action_handler is not None:
                    current_action_handler.process(action=current_action)

                time.sleep(self._config.sleep)

            for handler in self.processed_actions.values():
                handler.on_test_case_end()

            Context.clear()

    def close(self):
        self._common_factory.close()



import time

from run_one.action_handlers.abstract_action_handler import AbstractActionHandler
from run_one.util.util import Action


class SleepHandler(AbstractActionHandler):

    def process(self, action: Action):
        time.sleep(int(action.extra_data.get('Time', 0)) / 1000)



from th2_common.schema.grpc.router.grpc_router import GrpcRouter
from th2_common_utils import dict_to_message
from th2_grpc_act_template.act_service import ActService
from th2_grpc_act_template.act_template_pb2 import PlaceMessageRequest


from run_one.action_handlers.abstract_action_handler import AbstractActionHandler
from run_one.action_handlers.context import Context
from run_one.util.util import Action
from run_one.processors.th2_processor import Th2ProcessorConfig


class TH2ActHandler(AbstractActionHandler):

    def __init__(self, config: Th2ProcessorConfig, grpc_router: GrpcRouter) -> None:
        self._config = config
        self._act_service: ActService = grpc_router.get_service(ActService)  # type: ignore

    def _get_act_method(self, message_type: str):
        return {'NewOrderSingle': self._act_service.placeOrderFIX,
                'OrderCancelRequest': self._act_service.placeOrderCancelRequest,
                'OrderCancelReplaceRequest': self._act_service.placeOrderCancelReplaceRequest,
                'OrderMassCancelRequest': self._act_service.placeOrderMassCancelRequestFIX}[message_type]

    def process(self, action: Action):
        message_type = action.extra_data['MessageType']
        message = PlaceMessageRequest(message=dict_to_message(fields=action.row,
                                                              message_type=message_type,
                                                              session_alias=action.extra_data['User']))

        if parent_event_id := Context.get('parent_event_id'):
            message.parent_event_id.CopyFrom(parent_event_id)

        if not self._config.use_place_method:
            response = self._act_service.sendMessage(message)
        else:
            response = self._get_act_method(message_type)(message)

        Context.set('checkpoint_id', response.checkpoint_id)




from collections import defaultdict
from datetime import datetime
import json
from typing import Callable
import uuid

import pandas as pd

from run_one.util.config import Config


uid = str(uuid.uuid1())
counter = 0


class Action:

    def __init__(self, row: dict, extra_data: dict) -> None:
        self.row = row
        self.extra_data = extra_data

    def __repr__(self) -> str:
        return f'row={self.row}\nextra_data={self.extra_data}'


def generate_random_id(value: str = ''):
    global counter
    counter += 1
    if not value:
        return str(uuid.uuid1())
    return f'{uid[:len(value) - len(str(counter))]}{counter}'


def generate_time(time_format: str = '%Y-%m-%dT%H:%M:%S.%fZ'):
    return datetime.utcnow().strftime(time_format)


def read_csv_matrix(filepath: str,
                    config: Config,
                    id_function: Callable = generate_random_id,
                    time_function: Callable = generate_time) -> dict[str, list[Action]]:
    """
    Read matrix file
    :param str filepath: path to matrix file
    :param config: configuration class instance
    :param id_function: function to transform ID-like fields
    :param time_function: function to transform Time-like fields
    :return: Collection of filtered action_handlers combined by test cases
    """

    file = pd.read_csv(filepath)

    result = defaultdict(list)
    test_case_name = ''
    id_mapping, time_mapping = {}, {}

    for index, row in file.iterrows():

        seq = row.get('Seq')
        if seq == 'TEST_CASE_START':
            test_case_name = row.get('CaseName')
            continue
        elif seq == 'TEST_CASE_END':
            id_mapping.clear()
            time_mapping.clear()
            continue

        action = row.get('Action')

        if action in config.processed_actions:
            row.dropna(inplace=True)
            extracted_fields = row.get([field for field in config.fields_to_extract if field in row], [])
            row.drop(config.fields_to_drop + config.fields_to_extract, errors='ignore', inplace=True)
            row.rename(config.field_mapping, inplace=True)

            for field in config.nested_fields:
                if field in row and row[field] != '*':
                    row[field] = json.loads(row[field].replace("'", '"'))

            if id_function is not None:
                for field in config.regenerate_id_fields:
                    if field in row and row[field] != '*':
                        field_value = row[field]
                        if field_value.startswith('!='):
                            key = field_value[2:]
                            row[field] = f'!={id_mapping.setdefault(key, generate_random_id(key))}'
                        else:
                            row[field] = id_mapping.setdefault(field_value, generate_random_id(field_value))

            if time_function is not None:
                for field in config.regenerate_time_fields:
                    if field in row and row[field] != '*':
                        row[field] = time_mapping.setdefault(row[field], time_function())

            result[test_case_name].append(Action(row.to_dict(), extracted_fields.to_dict()))

    return result



from importlib import import_module
from typing import Type

from run_one.action_handlers.abstract_action_handler import AbstractActionHandler
from run_one.util.util import Action


class AbstractProcessor:

    def load_action_handlers(self, processed_actions_mapping: dict) -> dict[str, Type[AbstractActionHandler]]:
        result = {}
        for action, handler_name in processed_actions_mapping.items():
            try:
                action_handler_module = import_module(f'run_one.action_handlers.{handler_name}')
            except ModuleNotFoundError:
                try:
                    action_handler_module = import_module(f'action_handlers.{handler_name}')
                except ModuleNotFoundError:
                    raise ModuleNotFoundError(f'Action Handler class {handler_name} is not found')

            action_handler_class = getattr(action_handler_module, handler_name, None)

            if action_handler_class is None:
                raise ModuleNotFoundError(f'Action Handler class {handler_name} is not found')

            result[action] = action_handler_class

        return result

    def process(self, test_cases: dict[str, list[Action]]):
        """
        Process test cases
        :param test_cases: collection of test cases: test case name to list of its actions
        """
        pass



from th2_common.schema.grpc.router.grpc_router import GrpcRouter
from th2_common_utils.converters.filter_converters import FieldFilter, dict_to_root_message_filter
from th2_grpc_check1.check1_pb2 import CheckRuleRequest
from th2_grpc_check1.check1_service import Check1Service
from th2_grpc_common.common_pb2 import Direction, FilterOperation, ConnectionID

from run_one.action_handlers.abstract_action_handler import AbstractActionHandler
from run_one.action_handlers.context import Context
from run_one.processors.th2_processor import Th2ProcessorConfig
from run_one.util.util import Action


class TH2Check1Handler(AbstractActionHandler):

    def __init__(self, config: Th2ProcessorConfig, grpc_router: GrpcRouter) -> None:
        self._config = config
        self._check1_service: Check1Service = grpc_router.get_service(Check1Service)  # type: ignore

    def process(self, action: Action):
        message_type = action.extra_data['MessageType']
        user = action.extra_data['User']
        direction = getattr(Direction, action.extra_data.get('Direction', 'FIRST').upper())
        cl_ord_id = action.row.get("OrigClOrdID", action.row.get("ClOrdID", "none"))
        user_and_direction = (user, direction, cl_ord_id)

        message_filter = {}
        for k, v in action.row.items():
            if v == '*':
                message_filter[k] = FieldFilter(operation=FilterOperation.NOT_EMPTY)

            elif k in self._config.key_fields:
                is_not_equal_operation = v.startswith('!=')
                message_filter[k] = FieldFilter(
                    value=v[2:] if is_not_equal_operation else v,
                    key=True,
                    operation=FilterOperation.NOT_EQUAL if is_not_equal_operation else FilterOperation.EQUAL
                )

            else:
                value = (FieldFilter(value=v[2:], operation=FilterOperation.NOT_EQUAL)
                         if isinstance(v, str) and v.startswith('!=')
                         else v)
                message_filter[k] = value

        check_rule_request = CheckRuleRequest(
            connectivity_id=ConnectionID(session_alias=user),
            root_filter=dict_to_root_message_filter(message_type=message_type, message_filter=message_filter),
            # message_timeout=5000,
            timeout=int(action.extra_data.get('Time', 20000)),
            parent_event_id=Context.get('root_event_id'),
            description=f'{message_type} for {user} (ID: {action.extra_data["ID"]})',
            direction=direction,
            book_name=self._config.book)

        if parent_event_id := Context.get('parent_event_id'):
            check_rule_request.parent_event_id.CopyFrom(parent_event_id)

        if checkpoint_id := Context.get('checkpoint_id'):
            check_rule_request.checkpoint.CopyFrom(checkpoint_id)

        chain_id = Context.get('chain_id_mapping').get(user_and_direction)
        if chain_id is not None:
            check_rule_request.chain_id.CopyFrom(chain_id)

        response = self._check1_service.submitCheckRule(check_rule_request)
        if chain_id is None:
            Context.get('chain_id_mapping')[user_and_direction] = response.chain_id



import yaml


class Config:

    def __init__(self, **kwargs) -> None:

        self.matrix_path = '../test.csv'
        if 'matrix_path' in kwargs:
            self.matrix_path = kwargs['matrix_path']

        self.processed_actions = {'sleep': 'SleepHandler'}
        if 'processed_actions' in kwargs:
            self.processed_actions = kwargs['processed_actions']

        self.field_mapping = {}
        if 'field_mapping' in kwargs:
            self.field_mapping = kwargs['field_mapping']

        self.nested_fields = []
        if 'nested_fields' in kwargs:
            self.nested_fields = kwargs['nested_fields']

        self.fields_to_drop = []
        if 'fields_to_drop' in kwargs:
            self.fields_to_drop = kwargs['fields_to_drop']

        self.fields_to_extract = []
        if 'fields_to_extract' in kwargs:
            self.fields_to_extract = kwargs['fields_to_extract']

        self.regenerate_id_fields = []
        if 'regenerate_id_fields' in kwargs:
            self.regenerate_id_fields = kwargs['regenerate_id_fields']

        self.regenerate_time_fields = []
        if 'regenerate_time_fields' in kwargs:
            self.regenerate_time_fields = kwargs['regenerate_time_fields']

        if 'processor_config' in kwargs:
            self.processor_config = kwargs['processor_config']

    @staticmethod
    def read_config(filepath: str) -> 'Config':
        with open(filepath) as file:
            data = yaml.safe_load(file)

        return Config(**data)
